local M = {}

local partsCount = 0
local spacerCount = 0
local wheelSlotCount = 0

local outputPresetsPath = 'lua/ge/extensions/torsion/spacedOutJSON/outputPresets.json'
local spacerTemplatesPath = 'lua/ge/extensions/torsion/spacedOutJSON/spacer_templates.json'
local hubLUTPath = 'lua/ge/extensions/torsion/spacedOutJSON/hubLUT.json'

-- depreciated: overrides numrays from the rims
-- local insertObject = {numRays = 20}

local function safeDelete(filename)
  if not string.startswith(filename, '/vehicles/common/torsion/') or filename:find("%.%.") ~= nil then
    log('E', 'torsion_spacedOut', 'unable to delete file: not in appropriate folder: ' .. filename)
    return false
  end
  if not FS:fileExists(filename) and not FS:directoryExists(filename) then
    return true
  end
  if not FS:removeFile(filename) then
    local errMsg = FS:getLastError() or ''
    log('E', 'torsion_spacedOut', 'unable to delete file: ' .. filename .. ' : ' .. errMsg)
    return false
  end
  return true
end

-- this is the new function starting in v0.12 or something
-- doesn't seem to be any faster but I guess newer is better...
local function safeDeleteFolder(folderPath)
  -- delete old files in reverse order for empty folders before try to delete it
  local deleteFileList = FS:findFilesByRootPattern( folderPath, "*", -1, true, true )
  --dump(deleteFileList)
  for i = #deleteFileList, 1, -1 do
    safeDelete( deleteFileList[i] )
    --dump(deleteFileList[i])
  end
  safeDelete(folderPath)
end

-- this is the function from back in v0.10 or something
--[[
local function safeDeleteFolder(folderPath)
  -- delete old files in reverse order for empty folders before try to delete it
  local deleteFileList = FS:findFilesByRootPattern('game:' .. folderPath, "*", -1, true, true )
  --dump(deleteFileList)
  for i = #deleteFileList, 1, -1 do
    safeDelete( deleteFileList[i]:gsub('DIR:', '') )
	--print(deleteFileList[i]:gsub('DIR:', ''))
  end
end
]]--

local function loadDirectories(directories)
  --M.vehicleDirectory = directories[#directories] -- assumes the vehicle dir is always last
  --log('D', "jbeam.loadDirectories","set vehicle directory to "..tostring(M.vehicleDirectory))

  local allParts = {}
  log('D', "jbeam.loadDirectories","*** loading jbeam files ***")

  -- step 1: find all jbeam files
  local jbeamFiles = {}
  for _, directory in pairs(directories) do
    if not FS:directoryExists(directory) then
      log('W', "jbeam.loadDirectories", "error loading vehicle directory:"..directory.." / "..directory )
      goto continue
    end
    --log('D', "jbeam.loadDirectories","loading vehicle directory:"..directory)
    local folders = {}
    local basePath = "game:" .. directory
	print(basePath)
    local files = FS:findFilesByRootPattern(basePath, "*.jbeam", -1, true, false)
    for _, file in ipairs(files) do
      table.insert(jbeamFiles, file)
    end

    ::continue::
  end


  -- step 2: load them all from the Filesystem in one go - a lot more efficient to read all files first
  local fileContent = {}
  --dump(jbeamFiles)
  for k,v in pairs(jbeamFiles) do
    fileContent[v] = readFile(v)
  end

  -- step 2: load them all from the Filesystem in one go
  --perf.enable(1)
  local json = require("json")

  for k,v in pairs(jbeamFiles) do
    local content = fileContent[v]
    if content ~= nil then
      local editing = false
      local state, parts = pcall(json.decode, content, v, editing)
      if state == false then
        log('E', "jbeam.loadDirectories","unable to decode JSON: "..tostring(v))
        log('E', "jbeam.loadDirectories","JSON decoding error: "..tostring(parts))
        return nil
      end
      --log('D', "jbeam.loadDirectories","  * " .. v .. " - "..tableSize(parts).." parts")
      for i,v in pairs(parts) do
        if allParts[i] ~= nil then
          log('W', "jbeam.loadDirectories","Duplicated part: "..tostring(i))
        end
        allParts[i] = v;
      end
    else
      log('W', "jbeam.loadDirectories","unable to read file: "..v)
    end
  end

  --perf.disable()
  --perf.saveDataToCSV('json_performance.csv')

  -- no parts workaround :)
  if tableSize(allParts) == 1 then
    -- if there only one part, use that as main
    for partK,partV in pairs(allParts) do
      allParts[partK].slotType = "main"
    end
  end

  log('D', "jbeam.loadDirectories","*** jbeam loading done, found "..tableSize(allParts) .. ' parts in '..#jbeamFiles.. ' jbeam files')

return allParts
end

-- usage:
-- listSlots({'vehicles/common/tires/'})
local function listSlots(input)
	local inputParts = loadDirectories(input)
	--dump(inputParts)

	local outputData = {}
	
	for a,_ in pairs(inputParts) do 
		if inputParts[a].slotType ~= nil then
			local newSlot = inputParts[a].slotType
			local slotExists = 0
			for b,_ in pairs(outputData) do
				if outputData[b] == newSlot then
					slotExists = 1
				end
			end
			if slotExists == 0 then
				table.insert(outputData,newSlot)
			end
		end
	end
		
	return outputData
end

-- this is a function to adjust, for example, tireWidth.  tireWidth is inside pressureWheels so
-- we'd call the function like modifyTableOfArrays(outputFile[partName].pressureWheels, "tireWidth", -0.025)
local function modifyTableOfArraysAdd(container, target, modifier)
	for a,_ in pairs(container) do 
		if container[a][target] ~= nil then
			container[a][target] = container[a][target] + modifier
		end		
	end
	return container
end
local function modifyTableOfArraysMultiply(container, target, modifier)
	for a,_ in pairs(container) do 
		if container[a][target] ~= nil then
			container[a][target] = container[a][target] * modifier
		end		
	end
	return container
end


--generateOffsets({'vehicles/common/tires/'},"tires")
local function generateOffsets(partsPaths,partsLabel)

	local output = nil
	local slotsToMove = listSlots(partsPaths)  -- todo: I don't think this is used anywhere?
	local inputParts = loadDirectories(partsPaths)
	--dump(inputParts)
	local outputPresets = readJsonFile(outputPresetsPath)
	
	local inputTireCounter = 0
	local outputTireCounter = 0
	
	for offsetPart,_ in pairs(outputPresets) do
		local outputFile = {}
		for sourcePart,_ in pairs(inputParts) do
			local partName = tostring("torsion_" .. tostring(sourcePart) .. "_offset+" .. tostring(1000*outputPresets[offsetPart].offset))
			partsCount = partsCount + 1
			outputFile[partName] = deepcopy(inputParts[sourcePart])
			if outputFile[partName].information ~= nil then
        if outputFile[partName].information.authors ~= nil then
          outputFile[partName].information.authors = tostring(outputFile[partName].information.authors .. " [T]")
        end
      else
        --dumpToFile("badPart.json",outputFile[partName])
        --log if the loop has encountered a bad part
        local problematicSlotType = "[unknown]"
        if inputParts[sourcePart].slotType ~= nil then
          problematicSlotType = inputParts[sourcePart].slotType
        end
        log('E', 'torsion_spacedOut', 'JBEAM error with this part: ' .. tostring(sourcePart) .. ' for this slot: ' .. problematicSlotType .. '.  Missing "information" block will cause the game to crash when the part is loaded.')
      end
      
      if outputFile[partName].slotType ~= nil then
        outputFile[partName].slotType = tostring("torsion_" .. outputFile[partName].slotType .. "_offset+" .. tostring(1000*outputPresets[offsetPart].offset))
			else
        --dumpToFile("badPart.json",outputFile[partName])
        --log if the loop has encountered a bad part
        local problematicSlotType = "[unknown]"
        if inputParts[sourcePart].slotType ~= nil then
          problematicSlotType = inputParts[sourcePart].slotType
        end
        log('E', 'torsion_spacedOut', 'JBEAM error with this part: ' .. tostring(sourcePart) .. ' for this slot: ' .. problematicSlotType .. '.  Missing parent slot.')
      end
      
			-- this is where we adjust flexbody offset 
      if outputFile[partName].flexbodies ~= nil then 
        for a,_ in pairs(outputFile[partName].flexbodies) do
          if outputFile[partName].flexbodies[a][4] ~= nil then
            local posXOld = outputFile[partName].flexbodies[a][4].pos.x
            local xNeg = outputFile[partName].flexbodies[a][4].pos.x < 0
            outputFile[partName].flexbodies[a][4].pos.x = math.abs(outputFile[partName].flexbodies[a][4].pos.x) + outputPresets[offsetPart].offset
            if xNeg then outputFile[partName].flexbodies[a][4].pos.x = outputFile[partName].flexbodies[a][4].pos.x * -1 end
            --print(posXOld .. " -> " .. outputFile[partName].flexbodies[a][4].pos.x)
          end
        end
      else
        --dumpToFile("badPart.json",outputFile[partName])
        --log if the loop has encountered a bad part
        local problematicSlotType = "[unknown]"
        if inputParts[sourcePart].slotType ~= nil then
          problematicSlotType = inputParts[sourcePart].slotType
        end
        log('E', 'torsion_spacedOut', 'No flexbodies defined for this part: ' .. tostring(sourcePart) .. ' for this slot: ' .. problematicSlotType .. '.')
      end
			
			-- offset the slots
			if outputFile[partName].slots ~= nil then
				for a,_ in pairs(outputFile[partName].slots) do
					if outputFile[partName].slots[a][1] ~= "type" then
						outputFile[partName].slots[a][1] = tostring("torsion_" .. outputFile[partName].slots[a][1] .. "_offset+" .. tostring(1000*outputPresets[offsetPart].offset)) --slotType
						outputFile[partName].slots[a][2] = tostring("torsion_" .. outputFile[partName].slots[a][2] .. "_offset+" .. tostring(1000*outputPresets[offsetPart].offset)) --default part
					end
				end
			end
			
			if outputFile[partName].pressureWheels ~= nil then
				for a,_ in pairs(outputFile[partName].pressureWheels) do
					if outputFile[partName].pressureWheels[a].wheelOffset ~= nil then
						outputFile[partName].pressureWheels[a].wheelOffset = outputFile[partName].pressureWheels[a].wheelOffset - outputPresets[offsetPart].offset
					end
				end
			end
    end
		writeJsonFile(tostring('vehicles/common/torsion/spaced-out/torsion_spaced-out_' .. partsLabel .. "_" .. outputPresets[offsetPart].offset .. '.jbeam'), outputFile, true)
	end
	
end

local function generateSpacers()

	local partsPaths = {'vehicles/common/wheels/'}
	local outputPresets = readJsonFile(outputPresetsPath)
	local spacerTemplates = readJsonFile(spacerTemplatesPath)
	local hubLUT = readJsonFile(hubLUTPath)
	local inputParts = loadDirectories(partsPaths)
	local wheelSlots = listSlots(partsPaths)
	wheelSlotCount = table.getn(wheelSlots)
	local wheelSlotOffsets = {}
	local wheelSlotDefaults = {}
	
	-- This is not efficient or smart, I know.  Iterating over the same thing
	-- repeatedly is a waste.  Just going with what works for now.
	-- Pull the offsets from the existing wheels & grabbing a wheel to use as a default for the spacers.
	for wheelSlot,_ in pairs(wheelSlots) do
		for sourcePart,_ in pairs(inputParts) do
			for a,_ in pairs(inputParts[sourcePart].flexbodies) do
				if inputParts[sourcePart].slotType == wheelSlots[wheelSlot] then
					if inputParts[sourcePart].flexbodies[a][4] ~= nil then
						wheelSlotOffsets[wheelSlot] = math.abs(inputParts[sourcePart].flexbodies[a][4].pos.x)
						wheelSlotDefaults[wheelSlot] = tostring(sourcePart)
					end
				end
			end
		end
	end
	
	for wheelSlot,_ in pairs(wheelSlots) do
		local outputFile = {}
		local template = nil
		--determine whether this is a front or rear slot and use the correct template
		if string.find(wheelSlots[wheelSlot], "_R_") ~= nil then 
			template = deepcopy(spacerTemplates.spacer_R)
		else
			template = deepcopy(spacerTemplates.spacer_F)
		end
		for offsetPart,_ in pairs(outputPresets) do
			local partName = tostring("torsion_spacer_" .. tostring(wheelSlots[wheelSlot]) .. "_offset+" .. tostring(1000*outputPresets[offsetPart].offset))
			--print(partName)
			spacerCount = spacerCount + 1
			outputFile[partName] = deepcopy(template)
			outputFile[partName].slotType = tostring(wheelSlots[wheelSlot])
			--outputFile[partName].information.name = tostring("Spacers +" .. tostring(1000*outputPresets[offsetPart].offset) .. "mm")
			outputFile[partName].information.name = tostring(outputPresets[offsetPart].name)

			-- replace flexbodies
			for a,_ in pairs(outputFile[partName].flexbodies) do
				if outputFile[partName].flexbodies[a][1] ~= "mesh" then
					if outputPresets[offsetPart].flexbodyMesh ~= nil then
						outputFile[partName].flexbodies[a][1] = outputPresets[offsetPart].flexbodyMesh
					end
				end
			end
			
			-- this is where we set flexbody offset 
			local tempOffset
			if hubLUT[wheelSlots[wheelSlot]] ~= nil then 
				tempOffset = hubLUT[wheelSlots[wheelSlot]].offset
			else
				log('I', 'torsion_spacedOut', 'Offset for ' .. tostring(wheelSlot) .. ' not found in LUT, using guestimate.')
				tempOffset = wheelSlotOffsets[wheelSlot]
			end
			for a,_ in pairs(outputFile[partName].flexbodies) do
				if outputFile[partName].flexbodies[a][4] ~= nil then
					if outputFile[partName].flexbodies[a][4].rot.z == 180 then
						outputFile[partName].flexbodies[a][4].pos.x = (tempOffset) * -1
					else
						outputFile[partName].flexbodies[a][4].pos.x = tempOffset
					end
				end
			end
			
			-- set the slot
			for a,_ in pairs(outputFile[partName].slots) do
				if outputFile[partName].slots[a][1] ~= "type" then
					outputFile[partName].slots[a][1] = tostring("torsion_" .. wheelSlots[wheelSlot] .. "_offset+" .. tostring(1000*outputPresets[offsetPart].offset)) --slotType
					outputFile[partName].slots[a][2] = tostring("torsion_" .. wheelSlotDefaults[wheelSlot] .. "_offset+" .. tostring(1000*outputPresets[offsetPart].offset)) --default part
				end
			end
			
			if outputFile[partName].pressureWheels ~= nil then
				for a,_ in pairs(outputFile[partName].pressureWheels) do
					if outputFile[partName].pressureWheels[a].wheelOffset ~= nil then
						outputFile[partName].pressureWheels[a].wheelOffset = outputFile[partName].pressureWheels[a].wheelOffset - outputPresets[offsetPart].offset
					end
				end
			end
		end
		writeJsonFile(tostring('vehicles/common/torsion/spaced-out/torsion_spaced-out_' .. wheelSlots[wheelSlot] .. '_spacers' .. '.jbeam'), outputFile, true)		
	end
	

end

-- this function is wordy because it runs twice if it fails
-- if I get really crazy maybe I'll put it in a while loop, although that seems like a bad idea...
local function deleteTempFiles()
  local tempFilesPath = "/vehicles/common/torsion/spaced-out"
  log('I', 'torsion_spacedOut', 'Beginning to delete temporary parts.')
  local hp = hptimer()
  safeDeleteFolder(tempFilesPath)
  local hpOutput = hp:stop()/1000
  log('I', 'torsion_spacedOut', 'Temporary parts deleted. ' .. hpOutput .. 's')

  if FS:directoryExists(tempFilesPath) then
    log('E', 'torsion_spacedOut', 'Temp files not successfully deleted!  Retrying once...')
    log('I', 'torsion_spacedOut', 'Beginning to delete temporary parts.')
    local hp = hptimer()
    safeDeleteFolder(tempFilesPath)
    local hpOutput = hp:stop()/1000
    log('I', 'torsion_spacedOut', 'Temporary parts deleted. ' .. hpOutput .. 's')
  end
  
  if FS:directoryExists(tempFilesPath) then 
    log('E', 'torsion_spacedOut', 'Temp files not successfully deleted!  Failure.')
  end
end

local function generateAll()

	partsCount = 0
	spacerCount = 0
	wheelSlotCount = 0
	
	local hp = hptimer()
	
	if FS:fileExists(outputPresetsPath) and FS:fileExists(spacerTemplatesPath) and FS:fileExists(hubLUTPath) then 
		safeDeleteFolder("vehicles/common/torsion/spaced-out")
		generateOffsets({'vehicles/common/wheels/'},"wheels")
		generateOffsets({'vehicles/common/hubcaps/'},"hubcaps")
		generateOffsets({'vehicles/common/tires/'},"tires")
		generateSpacers()
		local hpOutput = round(hp:stop()/10)
		local hpOutput = hpOutput/100
		log('I', 'torsion_spacedOut', 'Ran for ' .. hpOutput .. ' seconds to generate a total of:')
		log('I', 'torsion_spacedOut', spacerCount .. ' spacers for ' .. wheelSlotCount .. ' front and rear wheel slots')
		log('I', 'torsion_spacedOut', partsCount .. ' additional parts to facilitate the spacer system')
	else
		log('I', 'torsion_spacedOut', 'Spaced Out was probably disabled using the mods manager.')
    deleteTempFiles()
	end
end

local function onExtensionLoaded()
	log('D', 'GELua.nodeStream.onExtensionLoaded', "Mods/Torsion's Spaced Out loaded")
	generateAll()
end

-- functions which should actually be exported
M.onExtensionLoaded = onExtensionLoaded
M.onModDeactivated = generateAll
M.onModActivated = generateAll
M.onExit = deleteTempFiles

--[[
-- troubleshooting only
M.safeDelete = safeDelete
M.safeDeleteFolder = safeDeleteFolder
M.generateOffsets = generateOffsets
M.listSlots = listSlots
M.generateSpacers = generateSpacers

-- working on parts cleanup after the mod is deleted or similar
local function modUnsubscribe()
  print("modUnsubscribe called")
end
M.modUnsubscribe = modUnsubscribe
]]
return M